# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from torch.optim.lr_scheduler import LambdaLR


class LinearLR(LambdaLR):
    """
    Learning rate scheduler with a learning rate that decreases linearly from the initial lr
    set in the optimizer to 0, after a warmup period during which it increases linearly from 0
    to the initial lr set in the optimizer.

    Args:
        optimizer:
            The optimizer for which to schedule the learning rate.
        num_warmup_steps:
            The number of steps for the warmup phase.
        num_training_steps:
            The total number of training steps.
        last_epoch:
            The index of the last epoch when resuming training.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        num_warmup_steps: int,
        num_training_steps: int,
        last_epoch: int = -1,
    ) -> None:
        self.num_warmup_steps = num_warmup_steps
        self.num_training_steps = num_training_steps
        super().__init__(optimizer, self._lr_lambda, last_epoch)

    def _lr_lambda(self, current_step: int) -> float:
        if current_step < self.num_warmup_steps:
            return float(current_step) / float(max(1, self.num_warmup_steps))
        return max(
            0.0,
            float(self.num_training_steps - current_step)
            / float(max(1, self.num_training_steps - self.num_warmup_steps)),
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import lightning.pytorch as pl
import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
from lightning.pytorch.strategies import DDPStrategy

from cellarium.ml.models.model import CellariumModel


class OnePassMeanVarStd(CellariumModel):
    """
    Calculate the mean, variance, and standard deviation of the data in one pass (epoch)
    using running sums and running squared sums.

    **References:**

    1. `Algorithms for calculating variance
       <https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance>`_.

    Args:
        g_genes:
            Number of genes.
        transform:
            If not ``None`` is used to transform the input data.
    """

    def __init__(self, g_genes: int, transform: nn.Module | None = None) -> None:
        super().__init__()
        self.g_genes = g_genes
        self.transform = transform
        self.x_sums: torch.Tensor
        self.x_squared_sums: torch.Tensor
        self.x_size: torch.Tensor
        self.register_buffer("x_sums", torch.zeros(g_genes))
        self.register_buffer("x_squared_sums", torch.zeros(g_genes))
        self.register_buffer("x_size", torch.tensor(0))
        self._dummy_param = torch.nn.Parameter(torch.tensor(0.0))

    @staticmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, np.ndarray | torch.Tensor]) -> tuple[tuple, dict]:
        x = tensor_dict["X"]
        return (x,), {}

    def forward(self, x_ng: torch.Tensor) -> None:
        if self.transform is not None:
            x_ng = self.transform(x_ng)

        self.x_sums = self.x_sums + x_ng.sum(dim=0)
        self.x_squared_sums = self.x_squared_sums + (x_ng**2).sum(dim=0)
        self.x_size = self.x_size + x_ng.shape[0]

    def on_train_start(self, trainer: pl.Trainer) -> None:
        if trainer.world_size > 1:
            assert isinstance(
                trainer.strategy, DDPStrategy
            ), "OnePassMeanVarStd requires that the trainer uses the DDP strategy."
            assert (
                trainer.strategy._ddp_kwargs["broadcast_buffers"] is False
            ), "OnePassMeanVarStd requires that broadcast_buffers is set to False."

    def on_epoch_end(self, trainer: pl.Trainer) -> None:
        # no need to merge if only one process
        if trainer.world_size == 1:
            return

        # merge the running sums
        dist.all_reduce(self.x_sums, op=dist.ReduceOp.SUM)
        dist.all_reduce(self.x_squared_sums, op=dist.ReduceOp.SUM)
        dist.all_reduce(self.x_size, op=dist.ReduceOp.SUM)

    @property
    def mean_g(self) -> torch.Tensor:
        return self.x_sums / self.x_size

    @property
    def var_g(self) -> torch.Tensor:
        return self.x_squared_sums / self.x_size - self.mean_g**2

    @property
    def std_g(self) -> torch.Tensor:
        return torch.sqrt(self.var_g)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
import os
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import pytest
import torch
from anndata import AnnData

from cellarium.ml import CellariumModule
from cellarium.ml.data import (
    DistributedAnnDataCollection,
    IterableDistributedAnnDataCollectionDataset,
)
from cellarium.ml.models import CellariumModel, GatherLayer
from cellarium.ml.utilities.data import AnnDataField, collate_fn, get_rank_and_num_replicas

# RuntimeError: Too many open files. Communication with the workers is no longer possible.
# Please increase the limit using `ulimit -n` in the shell or change the sharing strategy
# by calling `torch.multiprocessing.set_sharing_strategy('file_system')` at the beginning of your code
torch.multiprocessing.set_sharing_strategy("file_system")


class BoringModel(CellariumModel):
    """
    This model appends a batch input to an :attr:`iter_data` list at each iteration.
    Its intended use is for testing purposes where batch inputs can be inspected after
    iteration over the dataset with ``Trainer.fit()``. Batch input would typically contain
    feature counts, worker info, torch.distributed info, cache info, etc.
    """

    def __init__(self) -> None:
        super().__init__()
        self.iter_data: list = []
        self._dummy_param = torch.nn.Parameter(torch.tensor(0.0))

    @staticmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, np.ndarray | torch.Tensor]) -> tuple[tuple, dict]:
        return (), tensor_dict

    def forward(self, **batch: torch.Tensor) -> None:
        _, num_replicas = get_rank_and_num_replicas()
        if num_replicas > 1:
            for key, value in batch.items():
                batch[key] = torch.cat(GatherLayer.apply(value), dim=0)
        self.iter_data.append(batch)


@pytest.fixture(params=[[3, 6, 9, 12], [4, 8, 12], [4, 8, 11]])  # limits
def dadc(tmp_path: Path, request: pytest.FixtureRequest):
    limits = request.param
    n_cell = limits[-1]
    g_gene = 1

    X = np.arange(n_cell).reshape(n_cell, g_gene)
    adata = AnnData(X, dtype=X.dtype)
    for i, limit in enumerate(zip([0] + limits, limits)):
        sliced_adata = adata[slice(*limit)]
        sliced_adata.write(os.path.join(tmp_path, f"adata.00{i}.h5ad"))

    # distributed anndata
    filenames = str(os.path.join(tmp_path, f"adata.{{000..00{len(limits)-1}}}.h5ad"))
    dadc = DistributedAnnDataCollection(
        filenames,
        limits,
        max_cache_size=2,
        cache_size_strictly_enforced=True,
    )
    return dadc


@pytest.mark.parametrize("shuffle", [False, True], ids=["no shuffle", "shuffle"])
@pytest.mark.parametrize("num_workers", [0, 1, 2], ids=["zero workers", "one worker", "two workers"])
@pytest.mark.parametrize("batch_size", [1, 2, 3], ids=["batch size 1", "batch size 2", "batch size 3"])
def test_iterable_dataset(dadc: DistributedAnnDataCollection, shuffle: bool, num_workers: int, batch_size: int):
    n_obs = len(dadc)
    dataset = IterableDistributedAnnDataCollectionDataset(
        dadc,
        batch_keys={"X": AnnDataField("X")},
        batch_size=batch_size,
        shuffle=shuffle,
        test_mode=True,
    )
    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=collate_fn,
    )

    miss_counts = list(int(i) for batch in data_loader for i in batch["miss_count"])

    if num_workers > 1:
        worker_ids = list(int(i) for batch in data_loader for i in batch["worker_id"])
        for worker in range(num_workers):
            miss_count = max(c for c, w in zip(miss_counts, worker_ids) if w == worker)
            assert miss_count == math.ceil(len(dadc.limits) / num_workers)
    else:
        miss_count = max(miss_counts)
        assert miss_count == len(dadc.limits)

    actual_idx = list(int(i) for batch in data_loader for i in batch["X"])
    expected_idx = list(range(n_obs))

    # assert entire dataset is sampled
    assert len(expected_idx) == len(actual_idx)
    assert set(expected_idx) == set(actual_idx)


@pytest.mark.parametrize("shuffle", [False, True], ids=["no shuffle", "shuffle"])
@pytest.mark.parametrize("num_workers", [0, 1, 2], ids=["zero workers", "one worker", "two workers"])
@pytest.mark.parametrize("batch_size", [1, 2, 3], ids=["batch size 1", "batch size 2", "batch size 3"])
@pytest.mark.parametrize("drop_last", [False, True], ids=["no drop last", "drop last"])
def test_iterable_dataset_multi_device(
    dadc: DistributedAnnDataCollection,
    shuffle: bool,
    num_workers: int,
    batch_size: int,
    drop_last: bool,
):
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    n_obs = len(dadc)
    dataset = IterableDistributedAnnDataCollectionDataset(
        dadc,
        batch_keys={"X": AnnDataField("X")},
        batch_size=batch_size,
        shuffle=shuffle,
        drop_last=drop_last,
        test_mode=True,
    )
    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=collate_fn,
    )

    # fit
    model = BoringModel()
    module = CellariumModule(model)
    trainer = pl.Trainer(
        barebones=True,
        accelerator="cpu",
        devices=devices,
        max_epochs=1,  # one pass
    )
    trainer.fit(module, data_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    actual_idx = list(int(i) for batch in model.iter_data for i in batch["X"])
    expected_idx = list(range(n_obs))

    # assert entire dataset is sampled
    if drop_last and n_obs % devices != 0:
        expected_len = (n_obs // devices) * devices
        assert expected_len == len(actual_idx)
        assert set(actual_idx).issubset(expected_idx)
    else:
        expected_len = math.ceil(n_obs / devices) * devices
        assert expected_len == len(actual_idx)
        assert set(expected_idx) == set(actual_idx)


@pytest.mark.parametrize(
    "num_workers,persistent_workers",
    [(0, False), (1, False), (1, True), (2, False), (2, True)],
    ids=[
        "zero workers",
        "one not persistent worker",
        "one persistent worker",
        "two not persistent workers",
        "two persistent workers",
    ],
)
@pytest.mark.parametrize("epochs", [2, 3], ids=["two epochs", "three epochs"])
def test_iterable_dataset_set_epoch_multi_device(
    dadc: DistributedAnnDataCollection,
    num_workers: int,
    persistent_workers: bool,
    epochs: int,
):
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    dataset = IterableDistributedAnnDataCollectionDataset(
        dadc,
        batch_keys={"X": AnnDataField("X")},
        batch_size=1,
        shuffle=True,
        test_mode=True,
    )
    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=collate_fn,
        persistent_workers=persistent_workers,
    )

    # fit
    model = BoringModel()
    module = CellariumModule(model)
    trainer = pl.Trainer(
        barebones=True,
        accelerator="cpu",
        devices=devices,
        max_epochs=epochs,
        strategy="ddp",
    )
    trainer.fit(module, data_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    iter_data = model.iter_data

    actual_epochs = set(int(i) for batch in iter_data for i in batch["epoch"])
    expected_epochs = set(range(epochs))

    assert set(expected_epochs) == set(actual_epochs)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from jsonargparse import ArgumentParser, Namespace
from lightning.pytorch.utilities.rank_zero import rank_zero_warn

if TYPE_CHECKING:
    from cellarium.ml.core.module import CellariumModule


def _load_state(
    cls: type[CellariumModule],
    checkpoint: dict[str, Any],
    strict: bool | None = None,
    **cls_kwargs_new: Any,
) -> CellariumModule:
    """
    Re-implementation of :func:`lightning.pytorch.core.saving._load_state` that instantiates the model
    using the configuration saved in the checkpoint.
    """
    ### cellarium.ml - this part is different from the original
    parser = ArgumentParser()
    parser.add_class_arguments(cls, "model")

    cfg = Namespace(model={})

    # pass in the values we saved automatically
    if cls.CHECKPOINT_HYPER_PARAMS_KEY in checkpoint:
        cfg = checkpoint[cls.CHECKPOINT_HYPER_PARAMS_KEY]

    # update cfg with cls_kwargs_new, such that new has higher priority
    for key, value in cls_kwargs_new.items():
        cfg["model"][key] = value

    # instantiate the model
    cfg_init = parser.instantiate_classes(cfg)
    obj = cfg_init.model

    # save the cfg to the :attr:`obj.hparams` to be able to load the model checkpoint
    obj._set_hparams(cfg)
    ### cellarium.ml - end of modification

    # load the state_dict on the model automatically
    assert strict is not None
    keys = obj.load_state_dict(checkpoint["state_dict"], strict=strict)

    if not strict:
        if keys.missing_keys:
            rank_zero_warn(
                f"Found keys that are in the model state dict but not in the checkpoint: {keys.missing_keys}"
            )
        if keys.unexpected_keys:
            rank_zero_warn(
                f"Found keys that are not in the model state dict but in the checkpoint: {keys.unexpected_keys}"
            )

    return obj



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import logging
import warnings

import numpy as np
import pandas as pd
import torch


def get_highly_variable_genes(
    gene_names: list,
    mean: torch.Tensor,
    var: torch.Tensor,
    n_top_genes: int | None = None,
    min_disp: float | None = 0.5,
    max_disp: float | None = np.inf,
    min_mean: float | None = 0.0125,
    max_mean: float | None = 3,
    n_bins: int = 20,
) -> pd.DataFrame:
    r"""
    Get Highly variably genes. This is a replication of Highly Variable Genes from Scanpy with a ``Seurat`` flavor.

    **References:**

    1. `Highly Variable Genes from Scanpy
       <https://scanpy.readthedocs.io/en/stable/generated/scanpy.pp.highly_variable_genes.html>`_.

    Args:
        gene_names:
            Ensembl gene ids.
        mean:
            Gene expression means.
        var:
            Gene expression vars.
        n_top_genes:
            Number of highly-variable genes to keep.
        min_disp:
            If ``n_top_genes`` unequals None, this and all other cutoffs for the means and the normalized
            dispersions are ignored.
        max_disp:
            If ``n_top_genes`` unequals None, this and all other cutoffs for the means and the normalized
            dispersions are ignored.
        min_mean:
            If ``n_top_genes`` unequals None, this and all other cutoffs for the means and the normalized
            dispersions are ignored.
        max_mean:
            If ``n_top_genes`` unequals None, this and all other cutoffs for the means and the normalized
            dispersions are ignored.
        n_bins:
            Number of bins for binning the mean gene expression. Normalization is done with respect to each bin.
            If just a single gene falls into a bin, the normalized dispersion is artificially set to 1. You’ll be
            informed about this
    """
    # compute the dispersion
    assert len(gene_names) == len(mean) == len(var), "Sizes of `gene_names`, `mean`, and `var` should be the same"
    mean[mean == 0] = 1e-12  # set entries equal to zero to small value
    dispersion = var / mean
    # logarithmized mean as in Seurat
    dispersion[dispersion == 0] = np.nan
    dispersion = torch.log(dispersion)

    mean = torch.log1p(mean)
    # All the following quantities are "per-gene" here
    df = pd.DataFrame()
    df["means"] = mean
    df["dispersions"] = dispersion
    df["mean_bin"] = pd.cut(df["means"], bins=n_bins)
    disp_grouped = df.groupby("mean_bin")["dispersions"]
    disp_mean_bin = disp_grouped.mean()
    disp_std_bin = disp_grouped.std(ddof=1)
    df.index = gene_names
    # retrieve those genes that have nan std, these are the ones where
    # only a single gene fell in the bin and implicitly set them to have
    # a normalized disperion of 1
    one_gene_per_bin = disp_std_bin.isnull()
    gen_indices = np.where(one_gene_per_bin[df["mean_bin"].values])[0].tolist()
    if len(gen_indices) > 0:
        logging.debug(
            f"Gene indices {gen_indices} fell into a single bin: their "
            "normalized dispersion was set to 1.\n    "
            "Decreasing `n_bins` will likely avoid this effect."
        )
    # Circumvent pandas 0.23 bug. Both sides of the assignment have dtype==float32,
    # but there’s still a dtype error without “.value”.
    disp_std_bin[one_gene_per_bin.values] = disp_mean_bin[one_gene_per_bin.values].values
    disp_mean_bin[one_gene_per_bin.values] = 0
    # actually do the normalization
    df["dispersions_norm"] = (
        df["dispersions"].values - disp_mean_bin[df["mean_bin"].values].values  # use values here as index differs
    ) / disp_std_bin[df["mean_bin"].values].values

    dispersion_norm = df["dispersions_norm"].values
    if n_top_genes is not None:
        dispersion_norm = dispersion_norm[~np.isnan(dispersion_norm)]
        dispersion_norm[::-1].sort()  # interestingly, np.argpartition is slightly slower
        if n_top_genes > len(gene_names):
            logging.info("`n_top_genes` > `adata.n_var`, returning all genes.")
            n_top_genes = len(gene_names)
        if n_top_genes > len(dispersion_norm):
            warnings.warn(
                "`n_top_genes` > number of normalized dispersions, returning all genes with normalized dispersions.",
                UserWarning,
            )
            n_top_genes = len(dispersion_norm)
        disp_cut_off = dispersion_norm[n_top_genes - 1]
        gene_subset = np.nan_to_num(df["dispersions_norm"].values) >= disp_cut_off
        logging.debug(f"the {n_top_genes} top genes correspond to a " f"normalized dispersion cutoff of {disp_cut_off}")
    else:
        dispersion_norm[np.isnan(dispersion_norm)] = 0  # similar to Seurat
        gene_subset = np.logical_and.reduce(
            (
                mean > min_mean,
                mean < max_mean,
                dispersion_norm > min_disp,
                dispersion_norm < max_disp,
            )
        )

    df["highly_variable"] = gene_subset
    return df



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import os
from collections.abc import Sequence
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import pandas as pd
import torch


def write_prediction(
    prediction: torch.Tensor,
    ids: np.ndarray,
    output_dir: Path | str,
    postfix: int | str,
) -> None:
    """
    Write prediction to a CSV file.

    Args:
        prediction:
            The prediction to write.
        ids:
            The IDs of the cells.
        output_dir:
            The directory to write the prediction to.
        postfix:
            A postfix to add to the CSV file name.
    """
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=True)
    df = pd.DataFrame(prediction.cpu())
    df.insert(0, "db_ids", ids)
    output_path = os.path.join(output_dir, f"batch_{postfix}.csv")
    df.to_csv(output_path, header=False, index=False)


class PredictionWriter(pl.callbacks.BasePredictionWriter):
    """
    Write predictions to a CSV file. The CSV file will have the same number of rows as the
    number of predictions, and the number of columns will be the same as the prediction size.
    The first column will be the ID of each cell.

    .. note::
        To prevent an out-of-memory error, set the ``return_predictions`` argument of the
        :class:`~lightning.pytorch.Trainer` to ``False``.

    Args:
        output_dir:
            The directory to write the predictions to.
        prediction_size:
            The size of the prediction. If ``None``, the entire prediction will be
            written. If not ``None``, only the first ``prediction_size`` columns will be written.
    """

    def __init__(self, output_dir: Path | str, prediction_size: int | None = None) -> None:
        super().__init__(write_interval="batch")
        self.output_dir = output_dir
        self.prediction_size = prediction_size

    def write_on_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        prediction: torch.Tensor,
        batch_indices: Sequence[int] | None,
        batch: dict[str, np.ndarray | torch.Tensor],
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        if self.prediction_size is not None:
            prediction = prediction[:, : self.prediction_size]

        assert isinstance(batch["obs_names"], np.ndarray)
        write_prediction(
            prediction=prediction,
            ids=batch["obs_names"],
            output_dir=self.output_dir,
            postfix=batch_idx * trainer.world_size + trainer.global_rank,
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import lightning.pytorch as pl
import torch
from lightning.fabric.utilities.rank_zero import rank_zero_only

from cellarium.ml.models import ProbabilisticPCA


class VarianceMonitor(pl.Callback):
    r"""
    Automatically monitors and logs explained variance by the model during training.

    Args:
        total_variance: Total variance of the data. Used to calculate the explained variance ratio.
    """

    def __init__(self, total_variance: float | None = None):
        self.total_variance = total_variance

    def on_train_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        r"""
        Called when the train begins.

        Raises:
            AssertionError: If ``pl_module.model`` is not a ``ProbabilisticPCA`` instance.
            MisconfigurationException: If ``Trainer`` has no ``logger``.
        """
        assert isinstance(
            pl_module.model, ProbabilisticPCA
        ), "VarianceMonitor callback should only be used in conjunction with ProbabilisticPCA"

        if not trainer.loggers:
            raise pl.utilities.exceptions.MisconfigurationException(
                "Cannot use `LearningRateMonitor` callback with `Trainer` that has no logger."
            )
        # attempt to get the total variance from the checkpoint
        ppca = pl_module.model
        mean_var_std_ckpt_path = getattr(ppca, "mean_var_std_ckpt_path", None)
        if self.total_variance is None and mean_var_std_ckpt_path is not None:
            onepass = torch.load(mean_var_std_ckpt_path)
            self.total_variance = onepass.var_g.sum().item()

    @rank_zero_only
    def on_train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Called when the train batch ends."""
        assert isinstance(pl_module.model, ProbabilisticPCA)  # make mypy happy
        W_variance = pl_module.model.W_variance
        sigma_variance = pl_module.model.sigma_variance

        variance_stats = {}
        variance_stats["total_explained_variance"] = W_variance + sigma_variance
        variance_stats["W_variance"] = W_variance
        variance_stats["sigma_variance"] = sigma_variance
        if self.total_variance is not None:
            variance_stats["total_explained_variance_ratio"] = (W_variance + sigma_variance) / self.total_variance
            variance_stats["W_variance_ratio"] = W_variance / self.total_variance
            variance_stats["sigma_variance_ratio"] = sigma_variance / self.total_variance

        for logger in trainer.loggers:
            logger.log_metrics(variance_stats, step=trainer.fit_loop.epoch_loop._batches_that_stepped)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable, Sequence

import lightning.pytorch as pl
import numpy as np
import torch

from cellarium.ml.data import DistributedAnnDataCollection, IterableDistributedAnnDataCollectionDataset
from cellarium.ml.utilities.data import AnnDataField, collate_fn


class CellariumAnnDataDataModule(pl.LightningDataModule):
    """
    DataModule for :class:`~cellarium.ml.data.IterableDistributedAnnDataCollectionDataset`.

    Example::

        >>> from cellarium.ml import CellariumAnnDataDataModule
        >>> from cellarium.ml.utilities.data import AnnDataField, densify

        >>> dm = CellariumAnnDataDataModule(
        ...     "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...     shard_size=10_000,
        ...     max_cache_size=2,
        ...     batch_keys={
        ...         "x_ng": AnnDataField(attr="X", convert_fn=densify),
        ...         "feature_g": AnnDataField(attr="var_names"),
        ...     },
        ...     batch_size=5000,
        ...     shuffle=True,
        ...     seed=0,
        ...     drop_last=True,
        ...     num_workers=4,
        ... )
        >>> dm.setup()
        >>> for batch in dm.train_dataloader():
        ...     print(batch.keys())  # x_ng, feature_g

    Example::

        >>> from cellarium.ml import CellariumAnnDataDataModule
        >>> from cellarium.ml.utilities.data import AnnDataField, densify

        >>> dm = CellariumAnnDataDataModule(
        ...     "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...     shard_size=10_000,
        ...     max_cache_size=2,
        ...     batch_keys={
        ...         "x_ng": AnnDataField(attr="X", convert_fn=densify),
        ...         "feature_g": AnnDataField(attr="var_names"),
        ...     },
        ...     batch_size=5000,
        ...     shuffle=True,
        ...     seed=0,
        ...     drop_last=True,
        ...     num_workers=4,
        ... )
        >>> dm.setup()
        >>> for batch in dm.train_dataloader():
        ...     print(batch.keys())  # x_ng, feature_g

    Args:
        filenames:
            Names of anndata files.
        limits:
            List of global cell indices (limits) for the last cells in each shard.
            If ``None``, the limits are inferred from ``shard_size`` and ``last_shard_size``.
        shard_size:
            The number of cells in each anndata file (shard).
            Must be specified if the ``limits`` is not provided.
        last_shard_size:
            Last shard size. If not ``None``, the last shard will have this size possibly
            different from ``shard_size``.
        max_cache_size:
            Max size of the cache.
        cache_size_strictly_enforced:
            Assert that the number of retrieved anndatas is not more than maxsize.
        label:
            Column in :attr:`obs` to place batch information in. If it's ``None``, no column is added.
        keys:
            Names for each object being added. These values are used for column values for
            ``label`` or appended to the index if ``index_unique`` is not ``None``.
            If ``None``, ``keys`` are set to ``filenames``.
        index_unique:
            Whether to make the index unique by using the keys. If provided, this
            is the delimeter between ``{orig_idx}{index_unique}{key}``. When ``None``,
            the original indices are kept.
        indices_strict:
            If  ``True``, arrays from the subset objects will always have the same order
            of indices as in selection used to subset.
            This parameter can be set to ``False`` if the order in the returned arrays
            is not important, for example, when using them for stochastic gradient descent.
            In this case the performance of subsetting can be a bit better.
        batch_keys:
            Dictionary that specifies which attributes and keys of the :attr:`dadc` to return
            in the batch data and how to convert them. Keys must correspond to
            the input keys of the transforms or the model. Values must be instances of
            :class:`cellarium.ml.utilities.data.AnnDataField`.
        batch_size:
            How many samples per batch to load.
        shuffle:
            If ``True``, the data is reshuffled at every epoch.
        seed:
            Random seed used to shuffle the sampler if :attr:`shuffle=True`.
        drop_last:
            If ``True``, then the sampler will drop the tail of the data
            to make it evenly divisible across the number of replicas. If ``False``,
            the sampler will add extra indices to make the data evenly divisible across
            the replicas.
        test_mode:
            If ``True`` enables tracking of cache and worker informations.
        num_workers:
            How many subprocesses to use for data loading. ``0`` means that the data will be loaded in the main process.
    """

    def __init__(
        self,
        # DistributedAnnDataCollection args
        filenames: Sequence[str] | str,
        limits: Iterable[int] | None = None,
        shard_size: int | None = None,
        last_shard_size: int | None = None,
        max_cache_size: int = 1,
        cache_size_strictly_enforced: bool = True,
        label: str | None = None,
        keys: Sequence[str] | None = None,
        index_unique: str | None = None,
        indices_strict: bool = True,
        # IterableDistributedAnnDataCollectionDataset args
        batch_keys: dict[str, AnnDataField] | None = None,
        batch_size: int = 1,
        shuffle: bool = False,
        seed: int = 0,
        drop_last: bool = False,
        test_mode: bool = False,
        # DataLoader args
        num_workers: int = 0,
    ) -> None:
        super().__init__()
        # DistributedAnnDataCollection args
        self.filenames = filenames
        self.limits = limits
        self.shard_size = shard_size
        self.last_shard_size = last_shard_size
        self.max_cache_size = max_cache_size
        self.cache_size_strictly_enforced = cache_size_strictly_enforced
        self.label = label
        self.keys = keys
        self.index_unique = index_unique
        self.indices_strict = indices_strict
        # IterableDistributedAnnDataCollectionDataset args
        self.batch_keys = batch_keys or {}
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed
        self.drop_last = drop_last
        self.test_mode = test_mode
        # DataLoader args
        self.num_workers = num_workers
        # DistributedAnnDataCollection
        obs_columns = [field.obs_column for field in self.batch_keys.values() if field.obs_column is not None]
        self.dadc = DistributedAnnDataCollection(
            filenames=self.filenames,
            limits=self.limits,
            shard_size=self.shard_size,
            last_shard_size=self.last_shard_size,
            max_cache_size=self.max_cache_size,
            cache_size_strictly_enforced=self.cache_size_strictly_enforced,
            label=self.label,
            keys=self.keys,
            index_unique=self.index_unique,
            indices_strict=self.indices_strict,
            obs_columns=obs_columns,
        )

    @property
    def n_obs(self) -> int:
        return self.dadc.n_obs

    @property
    def n_vars(self) -> int:
        return self.dadc.n_vars

    @property
    def var_names(self) -> np.ndarray:
        return self.dadc.var_names.values

    def setup(self, stage: str | None = None) -> None:
        """
        .. note::
           setup is called from every process across all the nodes. Setting state here is recommended.

        """
        self.dataset = IterableDistributedAnnDataCollectionDataset(
            dadc=self.dadc,
            batch_keys=self.batch_keys,
            batch_size=self.batch_size,
            shuffle=self.shuffle,
            seed=self.seed,
            drop_last=self.drop_last,
            test_mode=self.test_mode,
        )

    def train_dataloader(self) -> torch.utils.data.DataLoader:
        """Training dataloader."""
        return torch.utils.data.DataLoader(
            self.dataset,
            num_workers=self.num_workers,
            collate_fn=collate_fn,
        )

    def predict_dataloader(self) -> torch.utils.data.DataLoader:
        """Prediction dataloader."""
        return torch.utils.data.DataLoader(
            self.dataset,
            num_workers=self.num_workers,
            collate_fn=collate_fn,
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

"""
Data utilities
--------------

This module contains helper functions for data loading and processing.
"""

import warnings
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import numpy as np
import pandas as pd
import scipy
import torch
import torch.distributed as dist
from anndata import AnnData
from anndata.experimental import AnnCollection
from torch.utils.data import get_worker_info as _get_worker_info


@dataclass
class AnnDataField:
    """
    Helper class for accessing fields of an AnnData-like object.

    Example::

        >>> from cellarium.ml.data import DistributedAnnDataCollection
        >>> from cellarium.ml.utilities.data import AnnDataField, densify

        >>> dadc = DistributedAnnDataCollection(
        ...     "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...     shard_size=10_000,
        ...     max_cache_size=2)

        >>> field_X = AnnDataField(attr="X", convert_fn=densify)
        >>> X = field_X(dadc)[:100]  # densify(dadc[:100].X)

        >>> field_cell_type = AnnDataField(attr="obs", key="cell_type")
        >>> cell_type = field_cell_type(dadc)[:100]  # np.asarray(dadc[:100].obs["cell_type"])

    Args:
        attr:
            The attribute of the AnnData-like object to access.
        key:
            The key of the attribute to access. If ``None``, the entire attribute is returned.
        convert_fn:
            A function to apply to the attribute before returning it.
            If ``None``, :func:`np.asarray` is used.
    """

    attr: str
    key: str | None = None
    convert_fn: Callable[[Any], np.ndarray] | None = None

    def __call__(self, adata: AnnData | AnnCollection) -> "AnnDataField":
        self.adata = adata
        return self

    def __getitem__(self, idx: int | list[int] | slice) -> np.ndarray:
        if self.adata is None:
            raise ValueError("Must call AnnDataField with an AnnData-like object first")

        value = getattr(self.adata[idx], self.attr)
        if self.key is not None:
            value = value[self.key]

        if self.convert_fn is not None:
            value = self.convert_fn(value)
        else:
            value = np.asarray(value)

        return value

    @property
    def obs_column(self) -> str | None:
        result = None
        if self.attr == "obs":
            result = self.key
        return result


def get_rank_and_num_replicas() -> tuple[int, int]:
    """
    This helper function returns the rank of the current process and
    the number of processes in the default process group. If distributed
    package is not available or default process group has not been initialized
    then it returns ``rank=0`` and ``num_replicas=1``.

    Returns:
        Tuple of ``rank`` and ``num_replicas``.
    """
    if not dist.is_available():
        num_replicas = 1
        rank = 0
    else:
        try:
            num_replicas = dist.get_world_size()
            rank = dist.get_rank()
        except RuntimeError:
            warnings.warn(
                "Distributed package is available but the default process group has not been initialized. "
                "Falling back to ``rank=0`` and ``num_replicas=1``.",
                UserWarning,
            )
            num_replicas = 1
            rank = 0
    if rank >= num_replicas or rank < 0:
        raise ValueError(f"Invalid rank {rank}, rank should be in the interval [0, {num_replicas-1}]")
    return rank, num_replicas


def get_worker_info() -> tuple[int, int]:
    """
    This helper function returns ``worker_id`` and ``num_workers``. If it is running
    in the main process then it returns ``worker_id=0`` and ``num_workers=1``.

    Returns:
        Tuple of ``worker_id`` and ``num_workers``.
    """
    worker_info = _get_worker_info()
    if worker_info is None:
        worker_id = 0
        num_workers = 1
    else:
        worker_id = worker_info.id
        num_workers = worker_info.num_workers
    return worker_id, num_workers


def collate_fn(batch: list[dict[str, np.ndarray]]) -> dict[str, np.ndarray | torch.Tensor]:
    """
    Collate function for the ``DataLoader``. This function assumes that the batch is a list of
    dictionaries, where each dictionary has the same keys. The values of each key are converted
    to a :class:`torch.Tensor` and concatenated along the first dimension. If the key is ``obs_names``,
    the values are concatenated along the first dimension without converting to a :class:`torch.Tensor`.

    Args:
        batch: List of dictionaries.

    Returns:
        Dictionary with the same keys as the input dictionaries, but with values concatenated along
        the batch dimension.
    """
    keys = batch[0].keys()
    collated_batch = {}
    if len(batch) > 1:
        assert all(keys == data.keys() for data in batch[1:]), "All dictionaries in the batch must have the same keys."
    for key in keys:
        if key == "obs_names":
            collated_batch[key] = np.concatenate([data[key] for data in batch], axis=0)
        elif key in ["var_names", "feature_g"]:
            # Check that all var_names are the same
            if len(batch) > 1:
                assert all(
                    np.array_equal(batch[0][key], data[key]) for data in batch[1:]
                ), "All dictionaries in the batch must have the same var_names."
            # If so, just take the first one
            collated_batch[key] = batch[0][key]
        else:
            collated_batch[key] = torch.cat([torch.from_numpy(data[key]) for data in batch], dim=0)
    return collated_batch


def densify(x: scipy.sparse.csr_matrix) -> np.ndarray:
    """
    Convert a sparse matrix to a dense matrix.

    Args:
        x: Sparse matrix.

    Returns:
        Dense matrix.
    """
    return x.toarray()


def categories_to_codes(x: pd.Series) -> np.ndarray:
    """
    Convert a pandas Series of categorical data to a numpy array of codes.
    Returned array is always a copy.

    Args:
        x: Pandas Index/Series/DataFrame object.

    Returns:
        Numpy array.
    """
    return np.asarray(x.cat.codes)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import pandas as pd
import pytest
from anndata import AnnData

from cellarium.ml.data import AnnDataSchema

n_cell_ref, n_cell, n_gene = (3, 5, 4)


@pytest.fixture
def ref_adata():
    X = np.ones((n_cell_ref, n_gene))
    L = np.ones((n_cell_ref, n_gene))
    M = np.ones((n_cell_ref, 2))
    V = np.zeros((n_gene, 2))
    P = np.zeros((n_gene, n_gene))
    obs = pd.DataFrame(
        {
            "A": np.array(["a", "b"])[np.random.randint(0, 2, n_cell_ref)],
            "B": np.array(["c", "d"])[np.random.randint(0, 2, n_cell_ref)],
            "C": pd.Categorical(
                np.array(["e", "f"])[np.random.randint(0, 2, n_cell_ref)],
                categories=["e", "f"],
            ),
        },
        index=[f"ref_cell{i:03d}" for i in range(n_cell_ref)],
    )
    var = pd.DataFrame(
        {
            "D": np.zeros(n_gene),
            "E": np.ones(n_gene),
            "F": np.arange(n_gene),
        },
        index=[f"gene{i:03d}" for i in range(n_gene)],
    )
    return AnnData(
        X,
        dtype=X.dtype,
        obs=obs,
        var=var,
        layers={"L": L},
        obsm={"M": M},
        varm={"V": V},
        varp={"P": P},
    )


@pytest.fixture
def adata():
    X = np.zeros((n_cell, n_gene))
    L = np.zeros((n_cell, n_gene))
    M = np.zeros((n_cell, 2))
    V = np.zeros((n_gene, 2))
    P = np.zeros((n_gene, n_gene))
    obs = pd.DataFrame(
        {
            "A": np.array(["a", "b"])[np.random.randint(0, 2, n_cell)],
            "B": np.array(["c", "d"])[np.random.randint(0, 2, n_cell)],
            "C": pd.Categorical(
                np.array(["e", "f"])[np.random.randint(0, 2, n_cell)],
                categories=["e", "f"],
            ),
        },
        index=[f"cell{i:03d}" for i in range(n_cell)],
    )
    var = pd.DataFrame(
        {
            "D": np.zeros(n_gene),
            "E": np.ones(n_gene),
            "F": np.arange(n_gene),
        },
        index=[f"gene{i:03d}" for i in range(n_gene)],
    )
    return AnnData(
        X,
        dtype=X.dtype,
        obs=obs,
        var=var,
        layers={"L": L},
        obsm={"M": M},
        varm={"V": V},
        varp={"P": P},
    )


@pytest.fixture
def schema(ref_adata: AnnData):
    return AnnDataSchema(ref_adata)


@pytest.mark.parametrize("delete_ref", [False, True])
def test_validate_adata(ref_adata: AnnData, adata: AnnData, delete_ref: bool):
    schema = AnnDataSchema(ref_adata)
    if delete_ref:
        del ref_adata
    schema.validate_anndata(adata)


@pytest.fixture(
    params=[
        "permute_obs_columns",
        "rename_obs_columns",
        "permute_var_columns",
        "change_obs_dtype",
        "change_obs_categories",
    ]
)
def change_adata(adata: AnnData, request: pytest.FixtureRequest):
    if request.param == "permute_obs_columns":
        adata.obs = adata.obs.iloc[:, [1, 0, 2]]
        err_msg = ".obs attribute columns for anndata passed in"

    elif request.param == "rename_obs_columns":
        cols = adata.obs.columns
        adata.obs = adata.obs.rename(columns={c: f"{c}_" for c in cols})
        err_msg = ".obs attribute columns for anndata passed in"

    elif request.param == "permute_var_columns":
        adata.var = adata.var.iloc[:, [1, 0, 2]]
        err_msg = ".var attribute for anndata passed in"

    elif request.param == "change_obs_dtype":
        adata.obs["A"] = np.ones(n_cell)
        err_msg = ".obs attribute dtypes for anndata passed in"

    elif request.param == "change_obs_categories":
        adata.obs["C"] = pd.Categorical(
            np.array(["g", "h"])[np.random.randint(0, 2, n_cell)],
            categories=["g", "h"],
        )
        err_msg = ".obs attribute dtypes for anndata passed in"

    return err_msg


def test_changed_adata(schema: AnnDataSchema, adata: AnnData, change_adata: str):
    err_msg = change_adata
    with pytest.raises(ValueError, match=err_msg):
        schema.validate_anndata(adata)


@pytest.mark.parametrize("change_obs_categories", [False, True])
@pytest.mark.parametrize("subset_obs_columns", [False, True])
def test_validate_obs_columns(
    ref_adata: AnnData,
    adata: AnnData,
    change_obs_categories: bool,
    subset_obs_columns: bool,
):
    if change_obs_categories:
        adata.obs["C"] = pd.Categorical(
            np.array(["g", "h"])[np.random.randint(0, 2, n_cell)],
            categories=["g", "h"],
        )

    if subset_obs_columns:
        schema = AnnDataSchema(ref_adata, obs_columns=["A", "B"])
    else:
        schema = AnnDataSchema(ref_adata)

    if change_obs_categories and not subset_obs_columns:
        with pytest.raises(ValueError, match=".obs attribute dtypes for anndata passed in"):
            schema.validate_anndata(adata)
    else:
        schema.validate_anndata(adata)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause


import torch
from torch import nn


class Log1p(nn.Module):
    """
    Log1p transform gene counts.

    .. math::

        y_{ng} = \\log(1 + x_{ng})
    """

    def __init__(self) -> None:
        super().__init__()

    def forward(self, x_ng: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x_ng: Gene counts.

        Returns:
            Log1p transformed gene counts.
        """
        return torch.log1p(x_ng)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
import os
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import pytest
import torch
from anndata import AnnData
from lightning.pytorch.strategies import DDPStrategy

from cellarium.ml import CellariumModule
from cellarium.ml.data import DistributedAnnDataCollection, IterableDistributedAnnDataCollectionDataset
from cellarium.ml.models import OnePassMeanVarStd, OnePassMeanVarStdFromCLI
from cellarium.ml.transforms import Log1p, NormalizeTotal
from cellarium.ml.utilities.data import AnnDataField, collate_fn
from tests.common import BoringDataset


@pytest.fixture
def adata():
    n_cell, g_gene = 10, 5
    rng = np.random.default_rng(1465)
    X = rng.integers(10, size=(n_cell, g_gene))
    return AnnData(X, dtype=X.dtype)


@pytest.fixture
def dadc(adata: AnnData, tmp_path: Path):
    # save anndata files
    limits = [2, 5, 10]
    for i, limit in enumerate(zip([0] + limits, limits)):
        sliced_adata = adata[slice(*limit)]
        sliced_adata.write(os.path.join(tmp_path, f"adata.00{i}.h5ad"))

    # distributed anndata
    filenames = str(os.path.join(tmp_path, "adata.{000..002}.h5ad"))
    return DistributedAnnDataCollection(
        filenames,
        limits,
        max_cache_size=2,  # allow max_cache_size=2 for IterableDistributedAnnDataCollectionDataset
        cache_size_strictly_enforced=True,
    )


@pytest.mark.parametrize("shuffle", [False, True])
@pytest.mark.parametrize("num_workers", [0, 2])
@pytest.mark.parametrize("batch_size", [1, 2, 3])
def test_onepass_mean_var_std_multi_device(
    adata: AnnData,
    dadc: DistributedAnnDataCollection,
    shuffle: bool,
    num_workers: int,
    batch_size: int,
):
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    # prepare dataloader
    dataset = IterableDistributedAnnDataCollectionDataset(
        dadc,
        batch_keys={"X": AnnDataField("X")},
        batch_size=batch_size,
        shuffle=shuffle,
    )
    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=collate_fn,
    )
    transform = torch.nn.Sequential(NormalizeTotal(target_count=10_000), Log1p())

    # fit
    model = OnePassMeanVarStd(g_genes=dadc.n_vars, transform=transform)
    module = CellariumModule(model)
    strategy = DDPStrategy(broadcast_buffers=False) if devices > 1 else "auto"
    trainer = pl.Trainer(
        barebones=True,
        accelerator="cpu",
        devices=devices,
        max_epochs=1,  # one pass
        strategy=strategy,  # type: ignore[arg-type]
    )
    trainer.fit(module, train_dataloaders=data_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # actual mean, var, and std
    actual_mean = model.mean_g
    actual_var = model.var_g
    actual_std = model.std_g

    # expected mean, var, and std
    x = transform(torch.from_numpy(adata.X))
    expected_mean = torch.mean(x, dim=0)
    expected_var = torch.var(x, dim=0, unbiased=False)
    expected_std = torch.std(x, dim=0, unbiased=False)

    np.testing.assert_allclose(expected_mean, actual_mean, atol=1e-5)
    np.testing.assert_allclose(expected_var, actual_var, atol=1e-4)
    np.testing.assert_allclose(expected_std, actual_std, atol=1e-4)


def test_load_from_checkpoint_multi_device(tmp_path: Path):
    n, g = 3, 2
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    # dataloader
    train_loader = torch.utils.data.DataLoader(
        BoringDataset(np.arange(n * g).reshape(n, g)),
        collate_fn=collate_fn,
    )
    # model
    init_args = {"g_genes": g, "target_count": 10}
    model = OnePassMeanVarStdFromCLI(**init_args)
    config = {
        "model": {
            "model": {
                "class_path": "cellarium.ml.models.OnePassMeanVarStdFromCLI",
                "init_args": init_args,
            }
        }
    }
    module = CellariumModule(model, config=config)
    # trainer
    strategy = DDPStrategy(broadcast_buffers=False) if devices > 1 else "auto"
    trainer = pl.Trainer(
        accelerator="cpu",
        strategy=strategy,  # type: ignore[arg-type]
        devices=devices,
        max_epochs=1,
        default_root_dir=tmp_path,
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # load model from checkpoint
    ckpt_path = tmp_path / f"lightning_logs/version_0/checkpoints/epoch=0-step={math.ceil(n / devices)}.ckpt"
    assert ckpt_path.is_file()
    loaded_model: OnePassMeanVarStdFromCLI = CellariumModule.load_from_checkpoint(ckpt_path).model
    # assert
    assert isinstance(model.transform, torch.nn.Sequential) and len(model.transform) == 2
    assert isinstance(loaded_model.transform, torch.nn.Sequential) and len(loaded_model.transform) == 2
    assert isinstance(model.transform[0], NormalizeTotal)
    assert isinstance(loaded_model.transform[0], NormalizeTotal)
    assert model.transform[0].target_count == loaded_model.transform[0].target_count
    np.testing.assert_allclose(model.mean_g, loaded_model.mean_g)
    np.testing.assert_allclose(model.var_g, loaded_model.var_g)
    np.testing.assert_allclose(model.std_g, loaded_model.std_g)
